"""
Provides elements that make stuff happen.

.. moduleauthor: Johannes Brachem <jbrachem@posteo.de>
"""

from typing import Union
from typing import List
from uuid import uuid4

import cmarkgfm
from cmarkgfm.cmark import Options as cmarkgfmOptions
from emoji import emojize

from ..exceptions import AlfredError
from .._helper import inherit_kwargs

from .core import jinja_env
from .core import _Choice
from .core import Row
from .core import Element
from .misc import JavaScript
from .input import SingleChoiceButtons
from .input import SingleChoiceList
from .input import SelectPageList

@inherit_kwargs
class SubmittingButtons(SingleChoiceButtons):
    """
    SingleChoiceButtons that trigger submission of the current page
    on click.

    Args:
        {kwargs}

    Examples:
        Using submitting buttons together with the
        :class:`.HideNavigation` element::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class Demo(al.Page):
                name = "demo"

                def on_exp_access(self):

                    self += al.HideNavigation()
                    self += al.SubmittingButtons("choice1", "choice2", name="b1")

    """

    def __init__(self, *choice_labels, **kwargs):
        super().__init__(*choice_labels, **kwargs)

    def added_to_page(self, page):
        
        super().added_to_page(page)

        t = jinja_env.get_template("js/submittingbuttons.js.j2")
        js = t.render(name=self.name)

        page += JavaScript(code=js)

@inherit_kwargs
class JumpButtons(SingleChoiceButtons):
    """
    SingleChoiceButtons that trigger jumps to specific pages on click.

    Each button can target a different page.

    Args:
        *choice_labels: Tuples of the form ``("label", "target_name")``,
            where "label" is the text that will be displayed and
            "target_name" is the name of the page that the experiment
            will jump to on click of the button.

        {kwargs}

    Attributes:
        targets (str): List of target page names.

    Examples:
        Simple jump buttons::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class Demo1(al.Page):
                name = "demo1"

                def on_exp_access(self):
                    self += al.JumpButtons(
                        ("jump to demo2", "demo2"),
                        ("jump to demo3", "demo3"),
                        name="jump1"
                        )

            @exp.member
            class Demo2(al.Page):
                name = "demo2"

            @exp.member
            class Demo3(al.Page):
                name = "demo3"

        Jump buttons with target page taken from the input on a previous
        page, using the :meth:`.Page.on_first_show` hook::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class Demo1(al.Page):
                name = "demo1"

                def on_exp_access(self):
                    self += al.TextEntry(toplab="Enter target page name", name="text1")

            @exp.member
            class Demo2(al.Page):
                name = "demo2"

                def on_first_show(self):
                    target = self.exp.values.get("text1")
                    self += al.JumpButtons(("jump to target", target), name="jump1")

            @exp.member
            class Demo3(al.Page):
                name = "demo3"

    """

    #: JavaScript template for the code that submits the form on click
    js_template = jinja_env.get_template("js/jumpbuttons.js.j2")

    def __init__(self, *choice_labels, button_style: Union[str, list] = "btn-primary", **kwargs):
        super().__init__(*choice_labels, button_style=button_style, **kwargs)
        self.choice_labels, self.targets = map(list, zip(*choice_labels))

    def prepare_web_widget(self):
        
        super().prepare_web_widget()

        self._js_code = []

        for choice, target in zip(self.choices, self.targets):
            js = self.js_template.render(id=choice.id, target=target)
            self.add_js(js)

    def validate_data(self):
        if not self.should_be_shown:
            return True
        
        cond1 = bool(self.input) if self.force_input else True
        return cond1

@inherit_kwargs
class DynamicJumpButtons(JumpButtons):
    """
    JumpButtons, where the target pages depend on the input to
    other elements on the same page.

    Args:
        *choice_labels: Tuples of the form ``("label", "element_name")``,
            where "label" is the text that will be displayed and
            "element_name" is the name of the element on the *same page*,
            whose input value will be inserted as the name of the target
            page.

        {kwargs}

    Attributes:
        targets (str): List of target page names.

    Examples:
        ::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class Demo1(al.Page):
                name = "demo1"

                def on_exp_access(self):
                    self += al.TextEntry(toplab="Enter a target page", name="text1")
                    self += al.DynamicJumpButtons(("Jump", "text1"), name="jump1")

            @exp.member
            class Demo2(al.Page):
                name = "demo2"

            @exp.member
            class Demo3(al.Page):
                name = "demo3"

    """

    # Documented at :class:`.JumpButtons`
    js_template = jinja_env.get_template("js/dynamic_jumpbuttons.js.j2")

    def validate_data(self):
        
        return True
        # cond1 = bool(self.data) if self.force_input else True

        # cond2 = True
        # for target in self.targets:
        #     value = self.experiment.data.get(target, None)
        #     cond2 = value in self.experiment.root_section.all_pages
        #     if not cond2:
        #         break

        # cond2 = all([target in self.experiment.root_section.all_pages for target in self.targets])
        # return cond1 and cond2

@inherit_kwargs
class JumpList(Row):
    """
    Allows participants to select a page from a dropdown menu and jump
    to it.

    Args:
        scope: Can be 'exp', or a section name. If *scope* is 'exp', all
            pages in the experiment are listed as choices. If *scope* is
            a section name, all pages in that section are listed.
        include_self: Whether to include the current page in the list,
            if it is in scope.
        check_jumpto: If *True* (default), pages that cannot be jumped
            to will be marked as disabled options. The evaluation is
            based on the :attr:`.Section.allow_jumpto` attribute of
            each page's direct parent section (and only that section).
        check_jumpfrom: If *True*, the element will check, if the
            current page can be left via jump. If not, all pages in the
            list will be marked as disabled options.
        show_all_in_scope: If *True* (default), all pages in the scope
            will be shown, including those that cannot be jumped to.
        display_page_name: If *True*, the page name will be displayed
            in the select list. Defaults to *True*.
        label: Label to display on the jump button.
        button_style: Style of the jump button. See
            :class:`.SingleChoiceButtons` for more details on this
            argument.
        button_round_corners: Boolean, determining whether the button
            should have rounded corners.
        debugmode: Boolean switch, telling the JumpList whether it
            should operate in debug mode.
        
        {kwargs}

    Notes:
        Different from other input-type elements, the JumpList does not
        have to be named.

    Examples:
        Minimal example::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class Demo1(al.Page):
                name = "demo2"

                def on_exp_access(self):
                    self += al.JumpList()

            @exp.member
            class Demo2(al.Page):
                name = "demo2"

    """

    def __init__(
        self,
        scope: str = "exp",
        include_self: bool = False,
        check_jumpto: bool = True,
        check_jumpfrom: bool = True,
        show_all_in_scope: bool = True,
        label: str = "Jump",
        button_style: Union[str, list] = "btn-dark",
        button_round_corners: bool = False,
        debugmode: bool = False,
        display_page_name: bool = True,
        **kwargs,
    ):

        random_name = "jumplist_" + uuid4().hex
        name = kwargs.get("name", random_name)
        select_name = name + "_select"
        btn_name = name + "_btn"
        select = SelectPageList(
            scope=scope,
            include_self=include_self,
            name=select_name,
            check_jumpto=check_jumpto,
            check_jumpfrom=check_jumpfrom,
            show_all_in_scope=show_all_in_scope,
            display_page_name=display_page_name
        )
        btn = DynamicJumpButtons(
            (label, select_name),
            name=btn_name,
            button_style=button_style,
            button_round_corners=button_round_corners,
        )
        super().__init__(select, btn, **kwargs)

        self.layout.width_sm = [10, 2]
        self.debugmode = debugmode

    def prepare_web_widget(self):
        
        super().prepare_web_widget()

        if not self.page.section.allow_jumpfrom:
            # disable button (disabling the list is controlled via init arguments of the list)
            self.elements[1].disabled = True

        if self.debugmode:
            for el in self.elements:
                el.disabled = False


@inherit_kwargs
class Button(Element):
    """
    A button that triggers execution of a python function on click.
    
    Args:
        text (str): Button text

        func (callable): Python function to be called on button click.
            The function must take zero arguments, but it can be a 
            method of a class or instance.
        
        followup (str): What to do after the python function was called.
            Can take the following values: 
            
            - ``refresh`` submits and reloads the current page (default),
            - ``none`` does nothing,
            - ``forward`` submits the current page and moves forward,
            - ``backward`` submits the current page and moves backward,
            - ``jump>page_name`` submits the current page and triggers
              a jump to a page with the name 'page_name'
            - ``custom`` executes custom JavaScript. If you choose this 
              option, you must supply your custom JavaScript through the
              argument *custom_js*.
        
        submit_first (bool): If True, the current values of all input
            elements on the current page will be saved on button click,
            before *func* is called. This way, these values will be
            available in *func* through :attr:`.ExperimentSession.values`, 
            if func has access to the ExperimentSession object.
            See Example 3. Defaults to True.
        
        button_style (str): Can be used for quick color-styling, using
            Bootstraps default color keywords: btn-primary, btn-secondary,
            btn-success, btn-info, btn-warning, btn-danger, btn-light,
            btn-dark. You can also use the "outline" variant to get
            outlined buttons (eg. "btn-outline-secondary"). Advanced users can
            supply their own CSS classes for button-styling. Defaults
            to "btn-primary".
        
        button_round_corners (bool): A boolean switch to toggle whether 
            the button should be displayed with rounded corners. Defaults 
            to False. 
        
        button_block (bool): A boolean switch to toggle whether the button
            should take up all horizontal space that is available. Can be
            quite useful when arranging buttons in :class:`.Row`s. 
            Defaults to False.
        
        custom_js (str): Custom JavaScript to execute after the python
            function specified in *func* was called. Only takes effect,
            if *followup* is set to 'custom'.

        {kwargs}
    
    Notes:

        Note that by default, the Button will not save data itself. If
        you wish to save information about button clicks, you can utilize
        the :attr:`.ExperimentSession.adata` dictionary in the button's
        callable. See Example 4.

        .. warning:: This element is very powerful. Remember that with great
            power comes great responsibility. Be aware that the callable
            *func* will be executed *every time* the button is clicked.
    
    Examples:

        Example 1: A minimal example that will print some text to your terminal
        window on button click::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class Demo(al.Page):

                def on_exp_access(self):
                    self += al.Button("Demo Button", func=self.demo_function)
                
                def demo_function(self):
                    print("\\nThis is a demonstration")
                    print(self.exp.exp_id)
                    print("\\n")
        
        Example 2: An example with a jump after the function call::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class Demo(al.Page):

                def on_exp_access(self):
                    self += al.Button("Demo Button", func=self.demo_function, followup="jump>page3")
            
                def demo_function(self):
                    print("\\nThis is a demonstration")
                    print(self.exp.exp_id)
                    print("\\n")

            exp += al.Page(title="Page 2", name="page2")
            exp += al.Page(title="Page 3", name="page3")

        
        Example 3: An example that uses values entered on the current page::
            
            import alfred3 as al
            exp = al.Experiment()


            @exp.member
            class Demo(al.Page):
                
                def on_exp_access(self):
                    self += al.TextEntry(leftlab="Enter", name="entry1")
                    self += al.Button("Demo Button", func=self.demo_function)

                def demo_function(self):
                    print("\\nThis is a demonstration")
                    print(self.exp.exp_id)
                    print(self.exp.values["entry1"])
                    print("\\n")
        
        Example 4: An example that saves a count of how often the button
        was clicked::
            
            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class Demo(al.Page):
                
                def on_exp_access(self):
                    self += al.Button("Demo Button", func=self.demo_function)

                def demo_function(self):
                    print("\\nThis is a demonstration")
                    
                    if "demo_button" in self.exp.adata:
                        self.exp.adata["demo_button"] += 1
                    else:
                        self.exp.adata["demo_button"] = 1
                        
                    print("\\n")


    """
    element_template = jinja_env.get_template("html/ActionButtonElement.html.j2")
    js_template = jinja_env.get_template("js/actionbutton.js.j2")

    def __init__(
        self, 
        text: str, 
        func: callable, 
        followup: str = "refresh",
        submit_first: bool = True,
        button_style: str = "btn-primary",
        button_round_corners: bool = False, 
        button_block: bool = False,
        custom_js: str = "",
        **kwargs):
        super().__init__(**kwargs)
        
        self.func = func
        self.submit_first = submit_first
        self.followup = followup
        self.url = None
        self.text = text
        self.custom_js = custom_js

        if not followup in {"refresh", "forward", "backward", "none", "custom"} and not followup.startswith("jump>"):
            raise ValueError(f"{followup} is an inappropriate value for 'followup'.")

        if self.followup == "custom" and not self.custom_js:
            raise ValueError("If you set 'followup' to 'custom', you must specify custom Javascritp to run.")

        self.button_style = button_style
        self.button_round_corners = button_round_corners
        self.button_block = button_block
    
    @property
    def template_data(self):
        d = super().template_data
        text = emojize(self.text, use_aliases=True)
        text = cmarkgfm.github_flavored_markdown_to_html(text, options=cmarkgfmOptions.CMARK_OPT_UNSAFE)
        d["text"] = text
        d["button_block"] = "btn-block" if self.button_block else ""
        d["button_style"] = self.button_style
        return d
    
    def prepare_web_widget(self):
        self.url = self.exp.ui.add_callable(self.func)

        # Javascript part
        self._js_code = []
        d = {}
        d["url"] = self.url
        d["expurl"] = f"{self.exp.ui.basepath}/experiment"
        d["followup"] = self.followup
        d["custom_js"] = self.custom_js
        d["name"] = self.name
        d["submit_first"] = self.submit_first
        d["set_data_url"] = self.exp.ui.set_page_data_url
        js = self.js_template.render(d)
        self.add_js(js)

        # Round corners part
        if self.button_round_corners:
            self._css_code = []
            css = f"#{ self.name } {{border-radius: 1rem;}}"
            self.add_css(css)
    


# -*- coding:utf-8 -*-

"""
.. moduleauthor:: Paul Wiemann <paulwiemann@gmail.com>
"""


from builtins import object
import os.path
import logging
import re
from uuid import uuid4
from typing import List

from . import alfredlog
from ._helper import check_name
from ._helper import _DictObj
from .exceptions import AlfredError


class ExpMember:
    """
    Baseclass for sections and pages.

    Args:
        name (str): Name of the member. Must be unique throughout the
            experiment. Can be defined as a class attribute. When you
            write a page or section in *class style*, the name can be
            inferred from the class name and does not need to be
            defined explicitly again. If a name is defined explicitly
            as a class attribute, that name is used.
        title (str): Title of the member. Can be defined as a class 
            attribute.
        subtitle (str): Subtitle of the member. Can be defined as a 
            class attribute.
        showif (dict): A dictionary, which can be used to define a 
            simple set of conditions under which the member will be
            shown. The conditions take the form of
            key-value pairs, where each key is the name of an input 
            element in the experiment and the value is the required 
            input. Can be defined as a class attribute.
        vargs (dict): A dictionary that can be used to pass additional
            arguments to the page. The arguments are then available as
            an instance attribute. As a special feature, the instance
            attribute allows you to access the values of the dictionary
            not only via the usual square-bracket notation, but also
            via dot-notation. 

            This argument fulfills a similar function as **kwargs do
            sometimes, but it makes sure that user-defined additional
            arguments will not collide with inherited keywords arguments.
            
            Can be defined as a class attribute.
    """


    #: Unique name of the member.
    name: str = None

    #: If *True*, the member will spawn a logger that can be configured
    #: individually for each instance
    instance_log: bool = False

    #: Dictionary of conditions that must be fulfilled for the member
    #: to be shown. Works similar to :attr:`.Element.showif`
    showif: dict = None
    
    #: Name of the parent section. Used when a member is appended to 
    #: the :class:`.Experiment`. If *None*, a member will be appended
    #: to the "_content" section.
    parent_name = None

    def __init__(
        self,
        name: str = None,
        title: str = None,
        subtitle: str = None,
        showif: dict = None,
        vargs: dict = None,
    ):

        self.log = alfredlog.QueuedLoggingInterface(base_logger=__name__)
        
        self._should_be_shown = True

        self._experiment = None
        self._parent_section = None
        self._section = None
        
        self._title = None
        self._subtitle = None
        self._has_been_shown = False
        self._has_been_hidden = False

        # the following assignments allow for assignment via class variables
        # during subclasses, but override the attributes, if given as
        # init parameters
        if showif is not None:
            self.showif = showif
        
        self._vargs = None
        if vargs is not None:
            self._vargs = _DictObj(vargs)
        elif self.vargs is not None:
            self.vargs = _DictObj(self.vargs)

        if title is not None:
            self.title = title

        if subtitle is not None:
            self.subtitle = subtitle

        self._name_set_via = {}

        if name is not None:
            self._set_name(name, via="argument")
        elif self.name is not None:
            self._set_name(self.name, via="class")
        
        if self.name is None:
            raise ValueError(f"{type(self).__name__} must be named!")
    
    @property
    def vargs(self) -> dict:
        """
        A dictionary of additional arguments passed on the class upon
        initialization. Can be defined as a class attribute.

        As a special feature, you can use dot notation to access
        (but not to set) values on this dictionary.

        Examples:

            Example of definig the *vargs* as a class attribute on a page::

                import alfred3 as al
                exp = al.Experiment()

                @exp.member
                class Demo(al.Page):
                    vargs = {"variable_argument": "demo"}

                    def on_exp_access(self):
                        self += al.Text(self.vargs.variable_argument)
            
            Example of using the *vargs* in page instantiation::

                import alfred3 as al
                exp = al.Experiment()

                class TestPage(al.Page):
                    def on_exp_access(self):
                        self += al.Text(f"{self.vargs.test}")

                @exp.member
                class Test(al.Section):

                    def on_exp_access(self):
                        for i in range(3):
                            self += TestPage(name=f"p{i}", vargs={"test": 123})

        """
        return self._vargs
        
        
    def _set_name(self, name: str, via: str):
        """
        Helps organize the different ways a name can be set.

        The ways are:

        1. As a class variable when deriving a page as a new class
        2. As an init argument when instantiating a class

        If a name was set via class variable, the init argument 
        can override it.

        """

        check_name(name)
        self._uid = name
        self._tag = name
        self.name = name
        self._name_set_via[via] = self.name

        if len(self._name_set_via) > 1:
            msg = f"Name of {self} was set via multiple methods. Current winner: '{self.name}', set via {via}."
            self.log.debug(msg)
        
    def _evaluate_showif(self) -> List[bool]:
        """Checks the showif conditions that refer to previous pages.
        
        Returns:
            A list of booleans, indicating for each condition whether
            it is met or not.
        """

        if self.showif:
            conditions = []
            for name, condition in self.showif.items():
                
                # raise error if showif evaluates current page
                if name in self.all_input_elements:
                    raise AlfredError(f"Incorrent showif definition for {self}, using '{name}' (can't use a member's own elements).")

                val = self.exp.data_manager.flat_session_data[name]
                conditions.append(condition == val)

            return conditions
        else:
            return [True]
    
    @property
    def all_input_elements(self) -> dict:
        """
        Redefined by pages and section.
        """
        return {}

    @property
    def tag(self) -> str:
        """
        Alias for :attr:`.name`, included under this name for backwards
        compatibility. For new code, use *name*.
        """
        return self._tag

    @tag.setter
    def tag(self, tag):
        if not (tag is None or isinstance(tag, str) or isinstance(tag, str)):
            raise TypeError("tag must be an instance of str or unicode")
        if self._tag is not None:
            raise ValueError("you're not allowed to change a tag.")
        self._tag = tag

    @property
    def uid(self) -> str:
        """
        Alias for :attr:`.name`, included under this name for backwards
        compatibility. For new code, use *name*.
        """
        return self._uid
    
    def visible(self, attr: str) -> dict:
        """
        Returns the subset of members in the given attribute that should
        be shown.
        
        Args:
            attr (str): Name of an attribute of the class which returns
                a dictionary of members. For instance, sections allow
                'all_members', 'members', 'all_pages', 'pages', and 
                pages allow 'elements', 'all_elements', 'input_elements',
                'filled_input_elements'.
        """
        d = getattr(self, attr)
        return {name: value for name, value in d.items() if value.should_be_shown}

    @property
    def should_be_shown(self) -> bool:
        """
        bool: Returns True if should_be_shown is set to True (default) and all
        showif conditions return True.
        """
        cond1 = self._should_be_shown
        cond2 = all(self._evaluate_showif())
        return cond1 and cond2

    @should_be_shown.setter
    def should_be_shown(self, b):
        """
        sets should_be_shown to b.

        :type b: bool
        """
        if not isinstance(b, bool):
            raise TypeError("should_be_shown must be an instance of bool")
        self._should_be_shown = b

    @property
    def title(self) -> str:
        """
        Page title (str).
        """
        if self._title is not None:
            return self._title
        else:
            return ""

    @title.setter
    def title(self, title):
        self._title = title

    @property
    def subtitle(self) -> str:
        """
        Page subtitle (str).
        """
        return self._subtitle

    @subtitle.setter
    def subtitle(self, subtitle):
        self._subtitle = subtitle

    def added_to_experiment(self, exp):
        """
        Informs the member that it was added to an experiment session.

        After this method is called, the member has access to the
        experiment session through the attribute :attr:`.exp`
        """
        if not self.name:
            raise AlfredError(f"{type(self).__name__} must have a unique name.")
        self._check_name_uniqueness(exp)
        self._experiment = exp

    def _check_name_uniqueness(self, exp):

        if self.name in exp.root_section.all_updated_members:
            raise AlfredError(f"Name '{self.name}' is already present in the experiment.")

        # if self.name != list(self._name_set_via.values())[-1]:
        #     raise AlfredError(f"{self}: Name must not be changed after assignment.")

        if self.name in dir(exp):
            raise ValueError(
                (
                    "The experiment has an attribute of the same name as"
                    f"the page '{self}'. Please choose a different page name."
                )
            )
    
    @property
    def experiment(self):
        """
        The :class:`.ExperimentSession` to which this member belongs.
        """
        return self._experiment
    
    @property
    def exp(self):
        """
        The :class:`.ExperimentSession` to which this member belongs.
        """
        return self._experiment
    
    @exp.setter
    def exp(self, value):
        self._experiment = value
    
    @experiment.setter
    def experiment(self, value):
        self._experiment = value

    def added_to_section(self, section):
        """
        Informs the member that it was added to a section.
        """
        self._parent_section = section
        self._section = section

    @property
    def section(self):
        """
        The member's parent section.
        """
        return self._section

    @property
    def parent(self):
        """
        Alias for :attr:`.section`.
        """
        return self._parent_section
    
    def uptree(self) -> list:
        """
        List of the parent section and the grandparent sections (recursive).
        """
        out = []
        if self.parent:
            out += [self.parent]
            out += self.parent.uptree()
            return out
        else:
            return out

    @property
    def tree(self) -> str:
        """
        A single string, indicating the member's position in the experiment.
        Consist of the names of the parent section and all grandparent
        sections, separated by dots.
        """
        if not self.parent:
            return self.tag

        if self.parent.tree:
            return self.parent.tree + "." + self.tag
        else:
            return self.tag

    @property
    def short_tree(self) -> str:
        """
        Short version of :attr:`.tree`, without the ``_root._content``
        part that is the same for all members.
        """
        return self.tree.replace("_root._content.", "")




"""
Provides configuration handling for alfred3 experiments.

.. moduleauthor:: Johannes Brachem <jbrachem@posteo.de>
"""


import os
import platform
import sys
import importlib

from configparser import ConfigParser, SectionProxy
from pathlib import Path
from typing import Union

from . import files
from ._helper import inherit_kwargs

class ExperimentConfig(ConfigParser):
    """
    Provides basic functionality for alfred3 configuration.

    Args:
        expdir: Path to the experiment directory.
        config_objects: A list of dictionaries and/or strings with alfred
            configuration in ini format. Defaults to `None`.
        **kwargs, inline_comment_prefixes: Keyword arguments that are passed on to 
            :class:`configparser.ConfigParser`
    
    Notes:
        This is a child class of :class:`configparser.ConfigParser`, 
        which parses alfred configuration files and objects on intialization
        and behaves just like a usual `ConfigParser` afterwards, most
        importantly providing the methods ``get(section, option)``,
        ``getint(section, option)``, ``getfloat(section, option)``,
        and ``getboolean(section, option)`` for retrieving values from the
        parser instance. There are many more methods available, which are
        documented in the official ConfigParser `documentation`_.
    
    .. _documentation: https://docs.python.org/3/library/configparser.html#configparser.ConfigParser
    """

    #: Environment variable key that corresponds to a 
    #: full filepath (including filename) to an alfred configuration
    #: file.
    env_location = "ALFRED_CONFIG_FILE"

    #: Name of the general library-wide configuration files.
    global_config_name = "alfred.conf"

    #: Name of the experiment-specific configuration file.
    exp_config_name = "config.conf"

    def __init__(self, expdir: str = None, config_objects: list = None, inline_comment_prefixes: str = ("#"), **kwargs):
        super().__init__(inline_comment_prefixes=inline_comment_prefixes, **kwargs)

        self._config_objects = config_objects if config_objects is not None else []

        self._config_files = []
        self.expdir = Path(expdir) if expdir is not None else Path.cwd()
        self._parse_alfred_config()

    def _collect_config_files(self):
        """
        Collect user-defined config files from config locations.
        """

        files = []

        if platform.system() in ["Linux", "Darwin"]:
            files.append(Path("/etc/alfred").joinpath(self.global_config_name))

        files.append(Path.home().joinpath(self.global_config_name))
        files.append(os.getenv(self.env_location))

        if self.expdir:
            files.append(self.expdir.joinpath(self.exp_config_name))

        self._config_files = [str(p) for p in files if p is not None]

    def _parse_alfred_config(self):
        """
        Parse all config files from different locations.
        """

        with importlib.resources.path(files, self.global_config_name) as f:
            self.read(f, encoding="utf-8")

        self._collect_config_files()
        self.read(self._config_files, encoding="utf-8")

        for obj in self._config_objects:
            if not isinstance(obj, (str, dict)):
                raise TypeError("Config objects must be list of strings or dictionaries.")

            try:
                self.read_dict(obj)
            except AttributeError:
                self.read_string(obj)

    def as_dict(self) -> dict:
        """
        Converts the ConfigParser structure into a nested dict.

        Each section name is a first level key in the the dict, and the 
        key values of the section becomes the dict in the second level::

            {
                'section_name': {
                    'key': 'value'
                }
            }
        
        Returns:
            dict: A dictionary representation of the parser instance.
        
        """

        return {section_name: dict(self[section_name]) for section_name in self.sections()}

    def get_section(self, name: str) -> SectionProxy:
        """
        Returns a section of the parser, if it exists.

        Args:
            name: Name of the section you wish to retrieve.

        Returns:
            SectionProxy: A :class:`~configparser.SectionProxy` object.
                If the requested section does not exist, the method
                returns `None`.
        """

        try:
            return self[name]

        except KeyError:
            return None

    def combine_sections(self, *args):
        """
        Parses the given sections on top of each other and returns
        the resulting section.

        Example with section in dict notation for simplicity::

            config = ExperimentConfig(expdir=Path.cwd())

            # config["sec1"] = {"key": "value"}
            # config["sec2"] = {"key": "different value"}
            # config["sec3"] = {"new_key": "new_value"}

            config.combine_section("sec1", "sec2", "sec3")
            # Result: {"key": "different_value", "new_key": "new_value"}

        """

        parser = ConfigParser()
        for section in args:
            parser.read_dict({"section": dict(self[section])})

        return parser["section"]

class ExperimentSecrets(ExperimentConfig):
    """
    Provides functionality for parsing secret config, e.g. DB credentials.

    This class is used only to set the attributes :attr:`env_location`,
    :attr:`global_config_name`, and :attr:`exp_config_name`. Otherwise
    it behaves exactly like :class:`ExperimentConfig`.

    Args:

    """

    env_location = "ALFRED_SECRETS_FILE"
    global_config_name = "secrets.conf"
    exp_config_name = "secrets.conf"




"""Provides basic logging configuration for alfred experiments.

.. moduleauthor: Johannes Brachem <jbrachem@posteo.de>
"""

from builtins import object
import logging
import sys
import os
import traceback
import queue
import threading
import copy
import re
from pathlib import Path
from typing import Union

from .config import ExperimentConfig

def prepare_file_handler(filepath: Union[str, Path]) -> logging.FileHandler:
    """Returns a :class:`~logging.FileHandler` and creates the necessary
    directories on the fly, if needed.

    Args:
        filepath: Absolute path to the targeted logfile.
    """
    logpath = Path(filepath)

    if not logpath.is_absolute():
        raise ValueError("Value of filepath must be an absolute path.")

    if logpath.is_dir():
        raise ValueError("Value of filepath must point to a file.")

    logdir = logpath.parent
    logdir.mkdir(exist_ok=True)

    return logging.FileHandler(str(logpath))


def prepare_alfred_formatter(exp_id: str) -> logging.Formatter:
    """Returns a :class:`~logging.Formatter` with the standard alfred
    logging format, including experiment id.
    """

    formatter = logging.Formatter(
        ("%(asctime)s - %(name)s - %(levelname)s - " f"experiment id={exp_id} - %(message)s")
    )

    return formatter


def parse_level(level: str) -> int:
    """Parses level definitions in lower case strings and returns the
    approriate level attribute of the Python logging library.
    """
    lvl = level.upper()
    if lvl not in ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]:
        raise ValueError("log level must be debug, info, warning, error or critical")
    return getattr(logging, lvl)


class QueuedLoggingInterface:
    """Interface for handling logging in alfred3.

    This class provides access to two loggers: The base logger and
    the queue logger. The reasoning is the following:

    * We want logging messages to contain experiment information
        (exp_id and session_id) and we want logfiles to potentially be
        experiment specific.
    
    * The necessary information only becomes available to pages and elements
        once the page or its parent section get appended to the 
        experiment.
    
    * So, logged messages are collected in a queue and logged as soon 
        as the experiment becomes available.
    
    * In order to not lose information in between their queueing
        and their eventual logging with the queue logger, all messages
        get also logged immediately using the base logger to a general
        "admin" log (this is turned off through configuration
        in run.py for local experiments by default, to avoid doubling
        log messages in a single file).
    
    Args:
        base_logger: Name of the base logger.
        queue_logger: Name of the queue logger.

    Attributes:
        use_base_logger (bool): Indicates, whether the base logger 
            should be used. Defaults to *True*, if a *base_logger* is
            defined upon initialization and to *False*, if *base_logger*
            is *None*. You can use this to turn off base logger usage.
        
        queue_logger (logging.Logger): The queue logger object. You
            can acces the logger this way to apply advanced 
            configuration.
        
        session_id (str): Allows you to set a session ID that will be
            included in logged messages. (Defaults to "NA".)
    """

    def __init__(self, base_logger: str = None, queue_logger: str = None):
        """Constructor method."""
        self.use_base_logger = True if base_logger is not None else False
        self._base_logger = logging.getLogger(base_logger) if base_logger is not None else None
        self._base_logger_storage = None
        self._queue_logger = logging.getLogger(queue_logger) if queue_logger is not None else None
        self._queue = queue.Queue()
        self._level = None
        self.session_id = "NA"

    @property
    def queue_logger(self):
        return self._queue_logger
    
    @queue_logger.setter
    def queue_logger(self, logger):

        if self.queue_logger is not None:
            self.warning(
                (
                    "Queue logger already present. Overriding queue logger "
                    f"{self.queue_logger} with {logger}."
                )
            )
        
        self._queue_logger = logger

    def add_queue_logger(self, obj, module: str):
        name = self.loggername(obj, module)
        self.queue_logger = logging.getLogger(name)
        if "experiment" in obj.__dict__:
            self.session_id = obj.experiment.session_id
        elif "exp" in obj.__dict__:
            self.session_id = obj.exp.session_id
        else:
            self.session_id = obj.experiment.session_id
        self.log_queued_messages()

    def loggername(self, obj, module: str) -> str:
        """Returns a logger name for use in :class:`~alfred3.alfredlog.QueueLoggingInterface.

        The name has the following format::

            exp.exp_id.module_name.class_name.object_identifier
        
        The identifier is only added if the object has an attribute
        *instance_log* that is set to *True*. If the object
        has a *name* attribute, that is used as the identifier. Else,
        the object's *uid* is used.
        
        Args:
            obj: The object for which a logger name should be generated.
        """
        # remove "alfred3" from module name
        module_name = module.split(".")
        module_name.pop(0)

        name = []
        name.append("exp")
        if "experiment" in obj.__dict__:
            name.append(obj.experiment.exp_id)
        elif "exp" in obj.__dict__:
            name.append(obj.exp.exp_id)
        else:
            name.append(obj.experiment.exp_id)

        
        name.append(".".join(module_name))
        name.append(type(obj).__name__)

        if "instance_log" in obj.__dict__:

            try:
                if obj.instance_log:
                    try:
                        name.append(obj.name)
                    except AttributeError:
                        name.append(obj.uid)

            except AttributeError as e:
                self.debug(f"Suppressed exception: {e}")

        return ".".join(name)

    def _unpack_worker(self):
        while not self._queue.empty():
            level, msg = self._queue.get()
            lvl_logger = getattr(self.queue_logger, level)
            lvl_logger(msg)

    def log_queued_messages(self):
        """Logs messages in the queue using the queue logger."""
        if self.queue_logger is None:
            pass

        if self._level:
            self.queue_logger.setLevel(self._level)
        threading.Thread(target=self._unpack_worker, name="alfredlog").start()

    def setLevel(self, level: str):
        """Sets a level for the queue logger. Since the level will be
        stored, this can be done even before a queue logger is defined.
        """

        if level.upper() not in ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]:
            raise ValueError(
                "Level must be one of 'DEBUG', 'INFO', 'WARNING', 'ERROR', or 'CRITICAL'."
            )

        self._level = getattr(logging, level.upper())

        if self.queue_logger:
            self.queue_logger.setLevel(self._level)

    def _handle_msg(self, msg: str, level: str, *args, **kwargs):

        msg = f"session id={self.session_id} - " + msg

        if self.use_base_logger:
            getattr(self._base_logger, level)(msg, *args, **kwargs)

        if not self.queue_logger:
            self._queue.put((level, msg))
        else:
            logger_lvl = getattr(self.queue_logger, level)
            try:
                logger_lvl(msg, *args, **kwargs)
            except TypeError:
                logger_lvl(*args, **kwargs)

    def deactivate_base_logger(self):
        self._base_logger_storage = self._base_logger
        self._base_logger = None
        self.use_base_logger = False

    def activate_base_logger(self):
        if not self._base_logger:
            self._base_logger = self._base_logger_storage
        self.use_base_logger = True

    def log(self, level: str, msg: str, *args, **kwargs):
        self._handle_msg(msg, level, *args, **kwargs)

    def debug(self, msg: str, *args, **kwargs):
        self._handle_msg(msg, "debug", *args, **kwargs)

    def info(self, msg: str, *args, **kwargs):
        self._handle_msg(msg, "info", *args, **kwargs)

    def warning(self, msg: str, *args, **kwargs):
        self._handle_msg(msg, "warning", *args, **kwargs)

    def error(self, msg: str, *args, **kwargs):
        self._handle_msg(msg, "error", *args, **kwargs)

    def critical(self, msg: str, *args, **kwargs):
        self._handle_msg(msg, "critical", *args, **kwargs)

    def exception(self, msg: str, *args, **kwargs):
        self._handle_msg(msg, "exception", *args, **kwargs)


class BWCompatibleLogger(QueuedLoggingInterface):
    """Offers backwards compatibility for logging.
    
    This class accepts an *experiment* parameter in its logging methods.
    """

    def process_experiment(self, exp):
        self.queue_logger = logging.getLogger(f"exp.{exp.exp_id}.general")
        self.session_id = exp.session_id

    def log(self, level: str, msg: str, *args, **kwargs):
        for arg in args:
            try:
                self.process_experiment(arg)
                args.remove(arg)
            except AttributeError:
                pass

        try:
            exp = kwargs.pop("experiment")
            self.process_experiment(exp)
        except KeyError:
            pass

        try:
            super().log(level=level, msg=msg, *args, **kwargs)
        except TypeError:
            super().log(level=level, msg=msg)

    def debug(self, msg: str, *args, **kwargs):
        for arg in args:
            try:
                self.process_experiment(arg)
                args.remove(arg)
            except AttributeError:
                pass

        try:
            exp = kwargs.pop("experiment")
            self.process_experiment(exp)
        except KeyError:
            pass

        try:
            super().debug(msg=msg, *args, **kwargs)
        except TypeError:
            super().debug(msg=msg)

    def info(self, msg: str, *args, **kwargs):
        for arg in args:
            try:
                self.process_experiment(arg)
                args.remove(arg)
            except AttributeError:
                pass

        try:
            exp = kwargs.pop("experiment")
            self.process_experiment(exp)
        except KeyError:
            pass

        try:
            super().info(msg=msg, *args, **kwargs)
        except TypeError:
            super().info(msg=msg)

    def warning(self, msg: str, *args, **kwargs):
        for arg in args:
            try:
                self.process_experiment(arg)
                args.remove(arg)
            except AttributeError:
                pass

        try:
            exp = kwargs.pop("experiment")
            self.process_experiment(exp)
        except KeyError:
            pass

        try:
            super().warning(msg=msg, *args, **kwargs)
        except TypeError:
            super().warning(msg=msg)

    def error(self, msg: str, *args, **kwargs):
        for arg in args:
            try:
                self.process_experiment(arg)
                args.remove(arg)
            except AttributeError:
                pass

        try:
            exp = kwargs.pop("experiment")
            self.process_experiment(exp)
        except KeyError:
            pass

        try:
            super().error(msg=msg, *args, **kwargs)
        except TypeError:
            super().error(msg=msg)

    def critical(self, msg: str, *args, **kwargs):
        for arg in args:
            try:
                self.process_experiment(arg)
                args.remove(arg)
            except AttributeError:
                pass

        try:
            exp = kwargs.pop("experiment")
            self.process_experiment(exp)
        except KeyError:
            pass

        try:
            super().critical(msg=msg, *args, **kwargs)
        except TypeError:
            super().critical(msg=msg)

    def exception(self, msg: str, *args, **kwargs):
        for arg in args:
            try:
                self.process_experiment(arg)
                args.remove(arg)
            except AttributeError:
                pass

        try:
            exp = kwargs.pop("experiment")
            self.process_experiment(exp)
        except KeyError:
            pass

        try:
            super().exception(msg=msg, *args, **kwargs)
        except TypeError:
            super().exception(msg=msg)




"""
Provides elements that display content.

.. moduleauthor: Johannes Brachem <jbrachem@posteo.de>
"""

import io
import time

from datetime import datetime
from typing import Union
from pathlib import Path
from uuid import uuid4

from emoji import emojize
import cmarkgfm
from cmarkgfm.cmark import Options as cmarkgfmOptions

from .._helper import is_url
from .._helper import inherit_kwargs

from .core import jinja_env
from .core import Element
from .core import RowLayout
from .core import LabelledElement
from .input import SingleChoiceButtons


class VerticalSpace(Element):
    """
    The easiest way to add vertical space to a page.

    Args:
        space: Desired space in any unit that is understood by a CSS
            margin (e.g. em, px, cm). Include the unit (e.g. '1em').

    Examples:

        Example of vertical space added between two text elements::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class HelloWorld(al.Page):
                name = "hello_world"

                def on_exp_access(self):
                    self += al.Text("Element 1")
                    self += al.VerticalSpace("100px")
                    self += al.Text("Element 2")

    """

    def __init__(self, space: str = "1em"):
        """Constructor method."""
        super().__init__()
        self.space = space

    @property
    def web_widget(self):

        # documented at baseclass
        return f"<div class='vertical-space-element' style='margin-bottom: {self.space};'></div>"


@inherit_kwargs
class Html(Element):
    """
    Displays html code on a page.

    Args:
        html: Html to be displayed.
        path: Filepath to a file with html code (relative to the
            experiment directory).

        {kwargs}

    Notes:
        This works very similar to :class:`.Text`. The most notable
        difference is that the *Text* element expects markdown, and
        therefore generally renders input text in a ``<p>`` tag. This
        is not always desirable for custom html, because it adds a
        margin at the bottom of the text.

        The *Html* element renders neither markdown, nor emoji shortcodes.

    Examples:
        Adding a simple div to the experiment::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class HelloWorld(al.Page):
                name = "hello_world"

                def on_exp_access(self):
                    self += al.Html(html="<div id='mydiv'>Text in div</div>")

    """

    element_template = jinja_env.get_template("html/TextElement.html.j2")

    def __init__(
        self,
        html: str = None,
        path: Union[Path, str] = None,
        **element_args,
    ):

        """Constructor method."""
        super().__init__(**element_args)

        self.html_code = html if html is not None else ""
        self.path = path

        if self._html_code and self.path:
            raise ValueError("You can only specify one of 'html' and 'path'.")

    @property
    def html_code(self) -> str:
        """str: The element's html code"""
        if self.path:
            return self.experiment.subpath(self.path).read_text()
        else:
            return self._html_code

    @html_code.setter
    def html_code(self, html):
        self._html_code = html

    @property
    def template_data(self) -> dict:

        d = super().template_data
        d["text"] = self.html_code

        return d


@inherit_kwargs
class Text(Element):
    """
    Displays text.

    You can use `GitHub-flavored Markdown`_ syntax and common
    `emoji shortcodes`_ . Additionally, you can use html for
    advanced formatting.

    Args:
        text: Text to be displayed.
        path: Filepath to a textfile (relative to the experiment
            directory).
        width: Element width. Usage is the same as in
            :class:`Element`, but the Text element uses its own
            specific default, which ensures good readability in
            most cases on different screen sizes.
        emojize: If True (default), emoji shortcodes in the text will
            be converted to unicode (i.e. emojis will be displayed).
        {kwargs}

    Examples:
        A simple text element, including a 😊 (``:blush:``) emoji added
        to a page::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class HelloWorld(al.Page):
                name = "hello_world"

                def on_exp_access(self):
                    self += al.Text("This is text :blush:")


    .. _GitHub-flavored Markdown: https://guides.github.com/features/mastering-markdown/
    .. _emoji shortcodes: https://www.webfx.com/tools/emoji-cheat-sheet/
    """

    element_template = jinja_env.get_template("html/TextElement.html.j2")

    def __init__(
        self,
        text: str = None,
        path: Union[Path, str] = None,
        width: str = None,
        emojize: bool = True,
        render_markdown: bool = True,
        **kwargs,
    ):

        """Constructor method."""
        super().__init__(width=width, **kwargs)

        self._text = text if text is not None else ""

        #: pathlib.Path: Path to a textfile, if specified in the init
        self.path: Path = Path(path) if path is not None else path

        #: bool: Boolean flag, indicating whether emoji shortcodes should be
        #: interpreted
        self.emojize: bool = emojize
        self.render_markdown: bool = render_markdown

        if self._text and self.path:
            raise ValueError("You can only specify one of 'text' and 'path'.")

    @property
    def text(self) -> str:
        """str: The text to be displayed"""
        if self.path:
            return self.experiment.subpath(self.path).read_text()
        else:
            return self._text

    def render_text(self) -> str:
        """
        Renders the markdown and emoji shortcodes in :attr:`.text`

        Returns:
            str: Text rendered to html code
        """

        text = self.text

        if self.emojize:
            text = emojize(text, use_aliases=True)
        if self.render_markdown:
            text = cmarkgfm.github_flavored_markdown_to_html(
                text, options=cmarkgfmOptions.CMARK_OPT_UNSAFE
            )
        return text

    @text.setter
    def text(self, text):
        self._text = text

    @property
    def element_width(self) -> str:

        if self.width is not None:
            return " ".join(self.converted_width)

        responsive = self.experiment.config.getboolean("layout", "responsive", fallback=True)
        if responsive:
            if self._element_width is None:
                return " ".join(["col-12", "col-sm-11", "col-lg-10", "col-xl-9"])
            else:
                return " ".join(self._element_width)
        elif not responsive:
            return "col-9"

    @property
    def template_data(self) -> dict:

        d = super().template_data
        d["text"] = self.render_text()

        return d


@inherit_kwargs
class Image(Element):
    """
    Displays an image.

    Args:
        path: Path to the image. Can be relative to the experiment
            directory, or absolute.
        url: URL to the image.
        {kwargs}

    Notes:
        You can specify *either* a path, or a url, but not both.

    Examples:
        Minimal example::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class Demo(al.Page):
                name = "demo1"

                def on_exp_access(self):
                    pylogo = "https://upload.wikimedia.org/wikipedia/commons/thumb/f/f8/Python_logo_and_wordmark.svg/1920px-Python_logo_and_wordmark.svg.png"

                    self += al.Image(url=pylogo)

    """

    # Documented at :class:`.Element`
    element_template = jinja_env.get_template("html/ImageElement.html.j2")

    def __init__(self, path: Union[str, Path] = None, url: str = None, **kwargs):
        super().__init__(**kwargs)

        self.path = path
        if url is not None and not is_url(url):
            raise ValueError("Supplied value is not a valid url.")
        else:
            self.url = url

        if path and url:
            raise ValueError("You can only specify one of 'path' and 'url'.")

        self.src = None

    def added_to_experiment(self, experiment):
        """
        The image is added to the dict of static files, if a path is
        provided.

        :meta private: (documented at :class:`.Element`)
        """
        super().added_to_experiment(experiment)
        if self.path:
            p = self.experiment.subpath(self.path)
            url = self.experiment.ui.add_static_file(p)
            self.src = url
        else:
            self.src = self.url

    @property
    def template_data(self):

        d = super().template_data
        d["src"] = self.src
        return d


@inherit_kwargs
class Audio(Image):
    """
    Allows playing audio files.

    Args:
        path: Path to the audio file. Can be relative to the experiment
            directory, or absolute.
        url: URL to the audio file.
        controls: If *True*, alfred will display controls like a pause/
            play button for participants to use (default: *True*).
        autoplay: If *True*, the audio file will start to play
            automatically (default: *False*).
        loop: If *True*, the audio file will start playing from the
            beginning again, once the end is reached (default: *False*).
        {kwargs}

    Notes:
        You can specify *either* a path, or a url, but not both.

    Examples:
        Minimal example::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class Demo(al.Page):
                name = "demo1"

                def on_exp_access(self):
                    demo_audio = "https://file-examples-com.github.io/uploads/2017/11/file_example_MP3_1MG.mp3"

                    self += al.Audio(url=demo_audio)

    """

    # Documented at :class:`.Element`
    element_template = jinja_env.get_template("html/AudioElement.html.j2")

    def __init__(
        self,
        path: Union[str, Path] = None,
        url: str = None,
        controls: bool = True,
        autoplay: bool = False,
        loop: bool = False,
        align: str = "center",
        **kwargs,
    ):
        super().__init__(path=path, url=url, align=align, **kwargs)
        self.controls = controls
        self.autoplay = autoplay
        self.loop = loop

    @property
    def template_data(self):

        d = super().template_data
        d["controls"] = self.controls
        d["autoplay"] = self.autoplay
        d["loop"] = self.loop

        return d


@inherit_kwargs
class Video(Audio):
    """
    Displays a video on the page.

    Args:
        path: Path to the video (relative to the experiment)
        url: Url to the video
        allow_fullscreen: Boolean, indicating whether users can enable
            fullscreen mode.
        video_height: Video height in absolute pixels (without unit).
            Defaults to "auto".
        video_width: Video width in absolute pixels (without unit).
            Defaults to "100%". It is recommended to use leave this
            parameter at the default value and use the general element
            parameter *width* for setting the width.
        {kwargs}

    Notes:
        You can specify *either* a path, or a url, but not both.

    Examples:
        Minimal example::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class Demo(al.Page):
                name = "demo1"

                def on_exp_access(self):
                    demo_video = "https://file-examples-com.github.io/uploads/2017/04/file_example_MP4_480_1_5MG.mp4"

                    self += al.Video(url=demo_video)

    """

    # Documented at :class:`.Element`
    element_template = jinja_env.get_template("html/VideoElement.html.j2")

    def __init__(
        self,
        path: Union[str, Path] = None,
        url: str = None,
        allow_fullscreen: bool = True,
        video_height: str = "auto",
        video_width: str = "100%",
        **kwargs,
    ):
        super().__init__(path=path, url=url, **kwargs)
        self.video_height = video_height
        self.video_width = video_width
        self.allow_fullscreen = allow_fullscreen

    @property
    def template_data(self):

        d = super().template_data
        d["video_height"] = self.video_height
        d["video_width"] = self.video_width
        d["allow_fullscreen"] = self.allow_fullscreen

        return d


@inherit_kwargs
class MatPlot(Element):
    """
    Displays a :class:`matplotlib.figure.Figure` object.

    Args:
        fig (matplotlib.figure.Figure): The figure to display.
        {kwargs}

    Notes:
        When plotting in alfred, you need to use the Object-oriented
        matplotlib API
        (https://matplotlib.org/3.3.3/api/index.html#the-object-oriented-api).

    Examples:
        Minimal example::

            from matplotlib.figure import Figure
            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class Demo(al.Page):
                name = "demo1"

                def on_exp_access(self):

                    # build an example plot
                    fig = Figure()
                    ax = fig.add_subplot()
                    ax.plot(range(10))

                    # add plot to page
                    self += al.MatPlot(fig=fig)

    """

    # Documented at :class:`.Element`
    element_template = jinja_env.get_template("html/ImageElement.html.j2")

    def __init__(self, fig, align: str = "center", **kwargs):
        super().__init__(align=align, **kwargs)
        self.fig = fig
        self.src = None

    def prepare_web_widget(self):

        out = io.BytesIO()
        self.fig.savefig(out, format="svg")
        out.seek(0)
        self.src = self.exp.ui.add_dynamic_file(out, content_type="image/svg+xml")

    @property
    def template_data(self):

        d = super().template_data
        d["src"] = self.src
        return d


class Hline(Element):
    """
    A simple horizontal line.

    Examples:
        Two text elements, separated by a horizontal line::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class Demo(al.Page):
                name = "demo"

                def on_exp_access(self):
                    self += al.Text(text="Text 1")
                    self += al.Hline()
                    self += al.Text(text="Text 2")

    """

    def __init__(self):
        super().__init__()

    def render_inner_html(self, template_data: dict) -> str:
        """
        A redefined render_inner_html method still needs to accept the
        *template_data* argument, even if it does not use it.

        :meta private: (documented at :class:`.Element`)
        """
        return "<hr>"


@inherit_kwargs
class CodeBlock(Text):
    """
    A convenience element for displaying highlighted code.

    Args:
        text: The code to be displayed.
        path: path: Filepath to a textfile (relative to the experiment
            directory) from which to read code.
        lang: The programming language to highlight. Defaults
            to 'auto', which tries to auto-detect the right language.
            See https://prismjs.com/index.html#supported-languages
            for an overview of possible language codes. Note that
            we may currently not support all possible languages.
        {kwargs}

    Examples:
        ::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class Demo(al.Page):
                name = "demo"

                def on_exp_access(self):
                    self += al.CodeBlock(text="console.log('test');", lang="javascript")

    """

    def __init__(
        self,
        text: str = None,
        path: Union[Path, str] = None,
        lang: str = "auto",
        width: str = "full",
        **element_args,
    ):

        """Constructor method."""
        super().__init__(text=text, path=path, width=width, **element_args)
        self.lang = lang if lang is not None else ""

    @property
    def text(self):

        if self.path:
            text = self.experiment.subpath(self.path).read_text()

            code = f"```{self.lang}\n{text}\n```"
            return code
        else:
            code = f"```{self.lang}\n{self._text}\n```"
            return code


@inherit_kwargs
class Label(Text):
    """
    A utility class, serving as label for other elements.

    Args:
        text: Text to be displayed.
        width: Usage as in :class:`.Element`, with the same default ('full').
        {kwargs}

    """

    def __init__(self, text, width="full", **kwargs):
        super().__init__(text=text, width=width, **kwargs)

        #: RowLayout: Layouting facility for controlling the column
        #: breaks and vertical alignment of the label. Gets set by
        #: :class:`.LabelledElement` automatically.
        self.layout: RowLayout = None

        #: Tells the label which column of the :attr:`.layout` it is
        self.layout_col: int = None

    @property
    def col_breaks(self) -> str:
        """The label's breakpoints for diferent screen sizes."""
        return self.layout.col_breaks(self.layout_col)

    @property
    def vertical_alignment(self) -> str:
        """The label's vertical alignment"""
        return self.layout.valign_cols[self.layout_col]


@inherit_kwargs
class ProgressBar(LabelledElement):
    """
    Displays a progress bar.

    Args:
        progress (str, float, int): Can be either "auto", or a number
            between 0 and 100. If "auto", the progress is calculated
            from the current progress of the experiment. The exact
            calculation can be further refined with the arguments
            'progress_base', 'n_elements', and 'n_pages'.

            If a number is supplied, that number will be used as the
            progress to be displayed.

            Defaults to 'auto'.

        bar_height (str): Height of the progress bar. Supply a string with
            unit, e.g. "6px".

        show_text (bool): Indicates, whether the progress bar should include
            text with the current progress.

        striped (bool): Indicates, whether the progress bar shoulb be striped.

        style (str): Determines the color of the progress bar. Possible values
            are "primary", "secondary", "info", "success", "warning",
            "danger", "light", "dark".

        animated (bool): Determines, whether a striped progress bar should be
            equipped with an animation.

        round (bool): Determines, whether the corners of the progress bar
            should be round.

        progress_base (str): A string, specifying the unit to use as the
            basis upon which progress should be calculated. Can be either
            'pages_elements', 'pages', or 'elements'. Defaults to
            'pages_elements', in which case the number of pages and the
            number of input elements are added together to form the
            denominator in the fraction for calculating progress.

        n_elements (int): Manual specification of the number of input
            elements in the experiment. If 'None', the experiment will
            try to infer the number of elements automatically, which may
            not always find the correct result. Defaults to *None*.

        n_pages (int): Manual specification of the number of
            page in the experiment. If 'None', the experiment will
            try to infer the number of pages automatically, which may
            not always find the correct result. Defaults to *None*.

        {kwargs}

    See Also:
        See :attr:`.ExperimentSession.progress_bar` for more information
        on the experiment-wide progress bar.

    Notes:
        If the argument *show_text* is *True*, the text's appearance
        can be altered via CSS. It receives a class of
        ":attr:`.css_class_element`-text" and an id of
        ":attr:`.Element.name`-text". You can use the method :meth:`.add_css`
        to append fitting CSS to the bar (see examples).

    Examples:

        Overriding the default experiment-wide progress bar::

            import alfred3 as al
            exp = al.Experiment()

            @exp.setup
            def setup(exp_session):
                exp_session.progress_bar = al.ProgressBar(show_text=True, bar_height="15px")

            exp += al.Page(name="example_page")

        Adding a progress bar as an element to a page::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class Example(al.Page):
                name = "example_page"

                def on_exp_access(self):
                    self += al.ProgressBar()

        Altering the progress bar text's apperance, applied to the
        experiment-wide progress bar. Note that the experiment-wide
        progress bar *always* receives the name "*progress_bar_*"::

            import alfred3 as al
            exp = al.Experiment()

            @exp.setup
            def setup(exp):
                exp.progress_bar = al.ProgressBar(show_text=True, bar_height="15px")
                exp.progress_bar.add_css("#progress_bar_ {{font-size: 12pt;}}")

            exp += al.Page(name="example_page")

    """

    element_template = jinja_env.get_template("html/ProgressBarElement.html.j2")

    def __init__(
        self,
        progress: Union[str, float, int] = "auto",
        bar_height: str = "6px",
        show_text: bool = False,
        striped: bool = True,
        style: str = "primary",
        animated: bool = False,
        round_corners: bool = False,
        progress_base: str = "pages_elements",
        n_elements: int = None,
        n_pages: int = None,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self._progress = None
        if progress != "auto":
            self.progress = progress

        self._progress_setting = progress
        self._progress_base = progress_base
        self._n_elements = n_elements
        self._n_pages = n_pages
        self._bar_height: str = bar_height
        self._show_text: bool = show_text
        self._striped: bool = striped
        self._bar_style: str = style
        self._animated: bool = animated
        self._round_corners: bool = "border-radius: 0;" if round_corners == False else ""

    def added_to_experiment(self, exp):

        super().added_to_experiment(exp)

        css = f".progress#{self.name}  {{height: {self._bar_height}; {self._round_corners}}}"
        self.add_css(code=css)

    def _prepare_web_widget(self):
        self.prepare_web_widget()

        try:
            self._activate_showif_on_current_page()
        except AttributeError as e:
            # special treatment for experiment-wide progress bar
            # because that one only has an experiment, no page
            if self.page is None and self.name == "progress_bar_":
                pass
            else:
                raise e

    @property
    def progress(self) -> Union[int, float]:

        if self._progress:  # manually defined via element
            return self._progress

        elif self.exp.current_page.progress:  # manually defined via page
            return self.exp.current_page.progress

        else:  # calculate automatically
            exact_progress = (self.numerator / self.denominator) * 100

            if not self.experiment.finished and not self.experiment.aborted:

                hi_bounded = min(round(exact_progress, 1), 95)
                lo_bounded = max(hi_bounded, 1)
                return lo_bounded
            else:
                return 100

    @progress.setter
    def progress(self, value: Union[int, float]):
        try:
            assert isinstance(value, (int, float))
            assert 0 < value and value < 100
        except AssertionError:
            raise ValueError("Progress must be a number between 0 and 100.")
        self._progress = value

    @property
    def n_elements(self):
        """
        int: Number of elements.
        """
        if self._n_elements:
            n_el = self._n_elements
        else:
            n_el = 0
            for el in self.exp.root_section.all_input_elements.values():
                if el.should_be_shown:
                    n_el += 1
                elif el.showif or el.page.showif or el.section.showif:
                    n_el += 0.3
        return n_el

    @property
    def n_pages(self):
        """
        int: Number of pages.
        """
        if self._n_pages:
            n_pg = self._n_pages
        else:
            n_pg = len(self.experiment.root_section.visible("all_pages"))

        return n_pg

    @property
    def denominator(self) -> int:
        """
        int: Denominator for the fraction in calculating the progress.
        """
        if self._progress_base == "pages_elements":
            return self.n_elements + self.n_pages
        elif self._progress_base == "pages":
            return self.n_pages
        elif self._progress_base == "elements":
            return self.n_elements

    @property
    def numerator(self) -> int:
        """
        int: Numerator for the fraction in calculating the progress.
        """
        if self._progress_base == "pages_elements":
            shown_el = len(self.experiment.root_section.all_shown_input_elements)
            shown_pg = len(self.experiment.root_section.all_shown_pages)
            return shown_el + shown_pg

        elif self._progress_base == "pages":
            shown_pg = len(self.experiment.root_section.all_shown_pages)
            return shown_pg

        elif self._progress_base == "elements":
            shown_el = len(self.experiment.root_section.all_shown_input_elements)
            return shown_el

    @property
    def template_data(self):

        d = super().template_data
        d["progress"] = self.progress
        d["show_text"] = self._show_text
        d["bar_height"] = self._bar_height
        d["bar_style"] = f"bg-{self._bar_style}"
        d["striped"] = "progress-bar-striped" if self._striped else ""
        d["animated"] = "progress-bar-animated" if self._animated else ""
        return d


class Alert(Text):
    """
    Allows the display of customized alerts.

    Args:
        text: Alert text
        category: Affects the appearance of alerts.
            Values can be: "info", "success", "warning", "primary",
            "secondory", "dark", "light", "danger".
        dismiss: Boolean parameter. If "True", AlertElement can be
            dismissed by a click. If "False", AlertElement is not
            dismissible. Default = "False"
        **element_args: Keyword arguments passed to the parent class
            :class:`TextElement`. Accepted keyword arguments are: name,
            font_size, align, width, position, showif,
            instance_log.

    Examples:

        >>> import alfred3 as al
        >>> alert = al.AlertElement(text="Alert text", dismiss=True, name="al1")
        >>> alert
        Alert(name='al1')

    """

    element_template = jinja_env.get_template("html/AlertElement.html.j2")

    def __init__(
        self, text: str = "", category: str = "info", dismiss: bool = False, **element_args
    ):
        super().__init__(text=text, **element_args)
        self.category = category
        self.dismiss = dismiss

    @property
    def template_data(self):
        d = super().template_data
        d["category"] = self.category
        d["role"] = "alert"
        d["dismiss"] = self.dismiss

        return d


class ButtonLabels(SingleChoiceButtons):
    """
    Disabled buttons to use for labelling.

    Args:
        *choice_labels: Variable numbers of choice labels. See
            :class:`.ChoiceElement` for details.
        {kwargs}

    .. warning:: Keep in mind that a table-like layout that uses
        :class:`.ButtonLabels` or :class:`.BarLabels` to label
        choice buttons will break on very small screens! Such layouts
        are only feasible on medium screens upwards.


    Examples:
        Using button labels to label single choice buttons::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class Demo(al.Page):
                name = "demo"

                def on_exp_access(self):

                    self += al.ButtonLabels("label1", "label2")
                    self += al.SingleChoiceButtons("choice1", "choice2", name="b1")

    """

    def __init__(self, *choice_labels, **kwargs):
        name = f"{type(self).__name__}" + uuid4().hex
        super().__init__(*choice_labels, disabled=True, name=name, **kwargs)

    @property
    def data(self):

        return {}


@inherit_kwargs
class BarLabels(ButtonLabels):
    """
    Disabled button bar to use for labelling.

    Args:
        *choice_labels: Variable numbers of choice labels. See
            :class:`.ChoiceElement` for details.
        {kwargs}

    .. warning:: Keep in mind that a table-like layout that uses
        :class:`.ButtonLabels` or :class:`.BarLabels` to label
        choice buttons will break on very small screens! Such layouts
        are only feasible on medium screens upwards.

    Examples:
        Using button labels to label single choice buttons::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class Demo(al.Page):
                name = "demo"

                def on_exp_access(self):

                    self += al.BarLabels("label1", "label2")
                    self += al.SingleChoiceBar("choice1", "choice2", name="b1")

    """

    # Documented at :class:`.SingleChoiceButtons`
    button_group_class = "choice-button-bar"

    # Documented at :class:`.SingleChoiceButtons`
    button_toolbar = True


@inherit_kwargs
class CountUp(Element):
    """
    Displays a timer, counting up from 00:00:00 (hh:mm:ss).

    Args:
        end_after (int): Optional argument for specifying an end for
            the counter after a number of seconds. Defaults to *-1*, which
            will let the counter run indefinitely.
        end_msg (str): Text to be displayed in the counter's place upon
            expiration.
        {kwargs}

    Examples:

        ::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class Demo(al.Page):
                def on_exp_access(self):
                    self += al.CountUp(font_size="big", align="center")

    """

    counter_js = jinja_env.get_template("js/countup.js.j2")
    element_template = jinja_env.get_template("html/TextElement.html.j2")

    def __init__(self, end_after: int = -1, end_msg: str = "expired", **kwargs):
        super().__init__(**kwargs)

        self.end_after = end_after
        self.end_msg = end_msg

    def prepare_web_widget(self):
        self._js_code = []
        js = self.counter_js.render(name=self.name, end_after=self.end_after, end_msg=self.end_msg)
        self.add_js(js)


@inherit_kwargs
class CountDown(CountUp):
    """
    Displays a countdown, counting down from now (time of page loading)
    to a specified end.

    Args:
        end_after (int): The amount of seconds after which the countdown
            ends. Alternatively, you can specifiy a specific time with
            the alternative constructor :meth:`.tilldate`.
        end_msg (str): Text to be displayed in the countdown's place upon
            expiration.
        reset (bool): If *True*, the countdown will start anew every time
            the page is reopened, reloaded, or refreshed. Defaults to 
            *False*, i.e. the countdown will continue where it left off.
        {kwargs}

    Notes:
        The CountDown element offers two alternative constructors:
        :meth:`.tilltime` (construction from UNIX timestamp) and 
        :meth:`.tilldate` (construction from date and 24h time 
        representation).

    Examples:
        Countdown running 30 seconds::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class Demo(al.Page):
                def on_exp_access(self):
                    self += al.CounDown(end_after=30, font_size="big", align="center")

        Countdown running until a certain datetime is reached::

            import alfred3 as al
            exp = al.Experiment()

            @exp.member
            class Demo(al.Page):
                def on_exp_access(self):
                    self += al.CounDown.tilldate(
                        year=2031,
                        month=1,
                        day=31,
                        hour=12,
                        minute=30,
                        second=12,
                        font_size="big",
                        align="center"
                        )
    """

    counter_js = jinja_env.get_template("js/countdown.js.j2")
    element_template = jinja_env.get_template("html/TextElement.html.j2")

    def __init__(self, end_after: int, end_msg: str = "expired", reset: bool = False, **kwargs):
        super().__init__(**kwargs)

        self.end_after_original = end_after
        self.end_after = None
        self.end_msg = end_msg

        self.start_time = None
        self.reset = reset

    @classmethod
    def tilltime(cls, t: Union[int, float], **kwargs):
        """
        Alternative constructor for a countdown targeted at a specific
        unix timestamp.

        Args:
            t (int, float): Target-time in seconds since EPOCH.
            **kwargs: Further keyword arguments are passed on to the
                ordinary constructor, see :class:`.CountDown`.
        
        Examples:
            Countdown running until July 18th 2036, 13:20:00 is reached::

                import alfred3 as al
                exp = al.Experiment()

                @exp.member
                class Demo(al.Page):
                    def on_exp_access(self):
                        self += al.CounDown.tilltime(
                            t=2_100_000_000,
                            font_size="big",
                            align="center"
                            )

        """
        if "end_after" in kwargs:
            raise TypeError(
                "'end_after' is an invalid keyword argument for the 'tilltime' constructor."
            )

        now = time.time()
        diff = t - now
        return cls(end_after=diff, **kwargs)

    @classmethod
    def tilldate(
        cls,
        year: int = None,
        month: int = None,
        day: int = None,
        hour: int = None,
        minute: int = None,
        second: int = None,
        **kwargs,
    ):
        """
        Alternative constructor for a countdown targeted at a specific date.

        Args:
            year, month, day, hour, minute, second: Time units, all
                integers. Cannot have leading zeroes. It is fine to
                specific only, for example, the year and the month. The
                arguments will be passed on to :class:`datetime.datetime`
                to construct a time object.

            **kwargs: Further keyword arguments are passed on to the
                ordinary constructor, see :class:`.CountDown`.

        Examples:

            Countdown running until a certain datetime is reached::

                import alfred3 as al
                exp = al.Experiment()

                @exp.member
                class Demo(al.Page):
                    def on_exp_access(self):
                        self += al.CounDown.tilldate(
                            year=2031,
                            month=1,
                            day=31,
                            hour=12,
                            minute=30,
                            second=12,
                            font_size="big",
                            align="center"
                            )
        """
        if "end_after" in kwargs:
            raise TypeError(
                "'end_after' is an invalid keyword argument for the 'tilldate' constructor."
            )

        dargs = {}
        if year:
            dargs["year"] = year
        if month:
            dargs["month"] = month
        if day:
            dargs["day"] = day
        if hour:
            dargs["hour"] = hour
        if minute:
            dargs["minute"] = minute
        if second:
            dargs["second"] = second

        date = datetime(**dargs)
        now = datetime.now()
        diff = date - now

        return cls(end_after=diff.total_seconds(), **kwargs)

    def prepare_web_widget(self):
        if not self.start_time:
            self.start_time = time.time()
            self.end_after = self.end_after_original

        elif not self.reset:
            now = time.time()
            already_passed = now - self.start_time
            self.end_after = self.end_after_original - already_passed

        super().prepare_web_widget()



from builtins import map, object
from builtins import callable as builtins_callable
import logging
import traceback
import re
import os

from uuid import uuid4
from pathlib import Path

from flask import (
    Flask,
    send_file,
    redirect,
    url_for,
    abort,
    request,
    make_response,
    session,
    send_from_directory,
    jsonify
)
from uuid import uuid4
from .config import ExperimentConfig, ExperimentSecrets
from . import alfredlog


class Script:
    exp = None
    exp_session = None
    expdir = None
    config = None
    secrets = None


app = Flask(__name__)
script = Script()

@app.route("/start", methods=["GET", "POST"])
def start():

    logger = logging.getLogger(f"alfred3")
    logger.info("Starting experiment initialization.")

    # Try-except block for compatibility with alfred3 previous to v1.2.0
    # TODO: Remove try-except block in v2.0.0 (keep "try" part)
    # pylint: disable=unsubscriptable-object
    exp_id = script.config.get("metadata", "exp_id")
    session_id = script.config.get("metadata", "session_id")
    # session_id = uuid4().hex
    log = alfredlog.QueuedLoggingInterface("alfred3", f"exp.{exp_id}")
    log.session_id = session_id
    script.log = log

    try:
        script.exp_session = script.exp.create_session(session_id=session_id, config=script.config, secrets=script.secrets, **request.args)
    except Exception:
        script.log.exception("Expection during experiment generation.")
        abort(500)

    # start experiment
    try:
        script.exp_session._start()
    except Exception:
        log.exception("Exception during experiment startup.")
        abort(500)

    # Experiment startup message

    session["page_tokens"] = []

    try:
        return redirect(url_for("experiment"))
    except Exception:
        log.exception("Exception during experiment startup.")
        script.exp_session.abort(
            reason="error",
            title="Oops - Something went wrong",
            icon="mug-hot",
            msg="Sorry, there was an error on our side."
        )
        return redirect(url_for("experiment"))


@app.route("/experiment", methods=["GET", "POST"])
def experiment():
    try:
        if request.method == "POST":

            move = request.values.get("move", None)
            page_token = request.values.get("page_token", None)

            try:
                token_list = session["page_tokens"]
                token_list.remove(page_token)
                session["page_tokens"] = token_list
            except ValueError:
                return redirect(url_for("experiment"))

            data = request.values.to_dict()
            data.pop("move", None)
            data.pop("directjump", None)
            data.pop("par", None)
            data.pop("page_token", None)

            script.exp_session.movement_manager.current_page._set_data(data)

            if move is None and not data:
                pass
            elif move:
                script.exp_session.movement_manager.move(direction=move)
            else:
                abort(400)

            return redirect(url_for("experiment"))

        elif request.method == "GET":
            url_pagename = request.args.get("page", None) # https://basepath.de/experiment?page=name
            if url_pagename:
                script.exp_session.movement_manager.jump_by_name(name=url_pagename)

            page_token = str(uuid4())

            # this block extracts the list "page_tokens", if it exists in the session
            # it creates the list "page_tokens" as an empty list, if not. This is needed
            # for qt-wk experiments because they don't call the route /start
            try:
                token_list = session["page_tokens"]
            except KeyError:
                token_list = []

            token_list.append(page_token)
            session["page_tokens"] = token_list

            html = script.exp_session.user_interface_controller.render_html(page_token)
            resp = make_response(html)
            resp.cache_control.no_cache = True
            return resp
    except Exception:
        script.log.exception("Exception during experiment execution.")
        script.exp_session.abort(
            reason="error",
            title="Oops - Something went wrong",
            icon="mug-hot",
            msg="Sorry, there was an error on our side (500)."
        )
        html = script.exp_session.user_interface_controller.render_html(page_token)
        resp = make_response(html)
        resp.cache_control.no_cache = True
        return resp



@app.route("/staticfile/<identifier>")
def staticfile(identifier):
    path, content_type = script.exp_session.user_interface_controller.get_static_file(identifier)
    dirname, filename = os.path.split(path)
    resp = make_response(send_from_directory(dirname, filename, mimetype=content_type))
    return resp


@app.route("/dynamicfile/<identifier>")
def dynamicfile(identifier):
    strIO, content_type = script.exp_session.user_interface_controller.get_dynamic_file(identifier)
    resp = make_response(send_file(strIO, mimetype=content_type))
    resp.cache_control.no_cache = True
    return resp


@app.route("/callable/<identifier>", methods=["GET", "POST"])
def callable(identifier):
    f = script.exp_session.user_interface_controller.get_callable(identifier)
    
    if request.content_type == "application/json":
        values = request.get_json()
    else:
        values = request.values.to_dict()
    values.pop("_", None)
    rv = f(**values)
    if rv is not None:
        resp = jsonify(rv)
    else:
        resp = make_response(redirect(url_for("experiment")))
    resp.cache_control.no_cache = True
    return resp

# @app.route("/None")
# def none(): pass